% Reproduction of PDE barrier option pricing with local vol.
% Andersen and Brotherton-Ratcliffe 
% "The equity option volatility smile: an implicit finite-difference approach" Winter 1997/98

clc; clear;

% Black-Scholes call price
BSCall = @(s,K,r,q,v,T) s*exp(-q*T)*normcdf((log(s/K) + (r-q+v^2/2)*T)/v/sqrt(T)) - K*exp(-r*T)*normcdf((log(s/K) + (r-q+v^2/2)*T)/v/sqrt(T) - v*sqrt(T));

% Spot, r, q
S0 = 590;
r = 0.06;
q = 0.0262;
PutCall = 'C';

% Implied volatility: Rows are maturities, so transpose
IVt = [...
    190	168	133	113	102	97	120	142	169	200;
    177	155	138	125	109	103	100	114	130	150;
    172	157	144	133	118	104	100	101	108	124;
    171	159	149	137	127	113	106	103	100	110;
    171	159	150	138	128	115	107	103	99	108;
    169	160	151	142	133	124	119	113	107	102;
    169	161	153	145	137	130	126	119	115	111;
    168	161	155	149	143	137	133	128	124	123;
    168	162	157	152	148	143	139	135	130	128;
    168	164	159	154	151	148	144	140	139	132]./1000;
IV = IVt';

% Strikes
K = [501.50 531.00 560.50 590.00 619.50 649.00 678.50 708.00 767.00 826.00];
NK = length(K);

% Maturities T
T = [0.175 0.425 0.695 0.94 1 1.5 2 3 4 5];
NT = length(T);

% Local volatility.  Transpose so columns are maturities
LVt = [...
    0.2708    0.3092    0.1507    0.1183    0.0974    0.1073    0.2536    0.3423    0.2400    0.2881;
    0.2423    0.1806    0.1480    0.1617    0.0940    0.0993    0.0752    0.0791    0.0865    0.1006;
    0.2206    0.1913    0.1573    0.1704    0.1262    0.0847    0.1057    0.1011    0.0897    0.1294;
    0.1667    0.1684    0.2162    0.1288    0.1857    0.0976    0.1022    0.0951    0.0753    0.1037;
    0.1519    0.1595    0.2372    0.1292    0.1671    0.1011    0.0975    0.0896    0.0792    0.1097;
    0.1886    0.1816    0.1579    0.1438    0.1453    0.1000    0.1576    0.1048    0.1244    0.0866;
    0.1931    0.1771    0.1691    0.1536    0.1412    0.1047    0.2990    0.0843    0.1359    0.1371;
    0.1496    0.1546    0.1662    0.1498    0.1675    0.1058    0.2142    0.1035    0.1284    0.1693;
    0.1382    0.1558    0.1903    0.1250    0.2936    0.1192    0.1672    0.1172    0.1271    0.1334;
    0.2206    0.2206    0.1842    0.1068    0.1653    0.2047    0.1689    0.0926    0.3379    0.4687];
LV = LVt';

% Interpolate the implied and local volatility surface
intmethod = 'spline';
dT = (T(NT) - T(1))/NT/10;
dK = (K(NK) - K(1))/NK/10;
Ti = T(1):dT:T(NT);
Ki = K(1):dK:K(NK);
LVi = interp2(T,K',LV,Ti,Ki',intmethod);
IVi = interp2(T,K',IV,Ti,Ki',intmethod);

% Plot the local and implied vols
surf(Ki,Ti,LVi)
xlabel('Strike')
ylabel('Maturity')
hold on
mesh(Ki,Ti,IVi)
hold off

%% Closed prices for Table 2 of Andersen and Brotherton-Ratcliffe
Mat = 2;
for k=1:NK
    vol(k) = interp2(T,K',IV,Mat,K(k));
    ClosedPrice(k) = BSCall(S0,K(k),r,q,vol(k),Mat);
end

%% Build the grids
% Uniform grid for the stock price and time
GridType = 'NonUniform';
nS = 10;
nT = 10;
I = eye((nS+1)*(nS+1));

% Maturity grid -- common to both uniform and non-uniform grid types
Tmax = Mat;
dt = Tmax/(nT-1);
T = [0:dt:Mat];

%% Flat extrapolation of the local volatility surface
KI = [0 Ki 2*Ki(end)];
TI = [0 Ti];
TopRow = LVi(1,:);
BotRow = LVi(end,:);
LeftCol = [LVi(1,1); LVi(:,1); LVi(end,1)];
LVI = [TopRow; LVi; BotRow];
LVI = [LeftCol LVI];


%% Crank-Nicolson scheme for the constant vol case, Table 2
theta = 0.5;
% imethod = 'Inverse' slow method, uses LU decomposition for inv(A);
% imethod = 'Tridiagonal' fast method, exploits tridiagonal form of A
imethod = 'Tridiagonal';

for k=1:NK
    % Stock price grid limits
     Smax = 1.75*K(k);
    % Build the grid
    if strcmp(GridType,'Uniform')
        Smin = 0.75*K(k);
        ds = (Smax-Smin)/nS;
        S = Smin:ds:Smax;
    else
        c = K(k)/30;
        dz = 1/nS*(asinh((Smax-K(k))/c) - asinh(-K(k)/c));
        for i=1:nS+1;
            z(i) = asinh(-K(k)/c) + (i-1)*dz;
            S(i) = K(k) + c*sinh(z(i));
        end
    end
    NS = nS+1;
    % Set the local vol surface to a constant volatility
    [R C] = size(LVI);
    LVIc = vol(k).*ones(R,C);
    % Get the price and implied vol
    FDPrice(k) = WeightedMethodDOBarrier(theta,S0,K(k),r,q,S,T,LVIc,TI,KI,dt,GridType,[],[],imethod,intmethod);
    FDIV(k) = BisecBSV(PutCall,S0,K(k),r,q,Mat,0.001,3.0,FDPrice(k),1e-5,1e+5);
    PriceError(k) = FDPrice(k) - ClosedPrice(k);
    VolError(k)   = FDIV(k) - vol(k);
end

%% Table 2 of Andersen and Brotherton-Ratcliffe
fprintf('---------------------------------------------------------------------------\n')
fprintf('Table 2 of Andersen and Brotherton-Ratclife \n')
fprintf('%s grid \n',GridType)
fprintf('Using price grid of %4.0f time grid of %4.0f \n',NS, nT+1)
fprintf('Strike  ImpVol    BSPrice    BSPricePDE   ImpVolPDE   PriceError   VolError\n')
fprintf('---------------------------------------------------------------------------\n')
for k=1:NK
    fprintf('%5.0f %8.4f   %8.4f    %8.4f    %8.4f    %8.4f    %8.4f \n',...
        K(k),vol(k),ClosedPrice(k),FDPrice(k),FDIV(k),PriceError(k),VolError(k));
end
fprintf('---------------------------------------------------------------------------\n')

%% Down-and-out calls for Table 4 of Andersen and Brotherton-Ratcliffe
Barrier = [500:10:550 555 560 570];
BarType = 'DO';
GridType = 'NonUniform';
nS = 60;
nT = 250;
Strike = 590;
Rebate = 0;
TruePrice = [59.5867 57.7751 55.3933 52.2785 48.2554 43.0306 39.8444 36.2468 27.4257];

tic
for k=1:length(Barrier)
    % Stock price grid limits
    Smax = 1.75*Strike;
    % Build the grid
    if strcmp(GridType,'Uniform')
        Smin = 0.75*Strike;
        ds = (Smax-Smin)/nS;
        S = Smin:ds:Smax;
    else
        c = K(k)/10;
        dz = 1/nS*(asinh((Smax-Strike)/c) - asinh(-Strike/c));
        for i=1:nS+1;
            z(i) = asinh(-Strike/c) + (i-1)*dz;
            S(i) = Strike + c*sinh(z(i));
        end
    end
    % Align the grid with the barrier
    Imin = find(S<Barrier(k));
    Imax = find(S>Barrier(k));
    inc1 = Barrier(k) - S(Imin(end));
    inc2 = S(Imax(1)) - Barrier(k);
    if inc1<inc2
        S = S+inc1;
    else
        S = S-inc2;
    end
    NS = nS+1;
    % Get the price
    BarrierPrice(k)  = WeightedMethodDOBarrier(theta,S0,Strike,r,q,S,T,LVI,TI,KI,dt,GridType,BarType,Barrier(k),imethod,intmethod);
end
toc
DollarError = TruePrice - BarrierPrice;


%% Figure 6 of Andersen and Brotherton-Ratcliffe
%Barr = [530 555 570];
Barr = [540 550 555];
vol =  0.001:.001:.50;

for k=1:length(vol)
    for b=1:length(Barr)
        price(k,b) = EuropeanDownAndOutCall(S0,Strike,Barr(b),r,q,vol(k),Mat,Rebate);
    end
end

% Plot the result
plot(vol,price(:,1),'k--',vol,price(:,2),'k-',vol,price(:,3),'k:')
line([vol(1) vol(end)], [BarrierPrice(5) BarrierPrice(5)]);
line([vol(1) vol(end)], [BarrierPrice(6) BarrierPrice(6)]);
line([vol(1) vol(end)], [BarrierPrice(7) BarrierPrice(7)]);
legend('H = $540', 'H = $550', 'H = $555', 'Location','NorthWest')
ylabel('Option Price')
xlabel('Implied Volatility')


%% Implied volatilities
BarrierIV = zeros(length(Barrier),2);
Tol = 1e-5;
MaxIter = 1e5;
a = 0.01;
b = 2.0;

for k=1:length(Barrier)
    if Barrier(k) <= 530
        BarrierIV(k,1) = BarrierImpliedVolBisection(S0,Strike,r,q,Mat,a,b,BarrierPrice(k),Tol,MaxIter,Barrier(k),Rebate);
    end
     if Barrier(k) == 540
         start = 0.12;
         BarrierIV(k,1) = fminsearch(@(v) BarrierImpliedVolObjFun(v,S0,Strike,r,q,Mat,BarrierPrice(k),Barrier(k),Rebate), start);
         start = 1.6;
         BarrierIV(k,2) = fminsearch(@(v) BarrierImpliedVolObjFun(v,S0,Strike,r,q,Mat,BarrierPrice(k),Barrier(k),Rebate), start);
     end
    if Barrier(k) == 550
        start = 0.1;
        BarrierIV(k,1) = fminsearch(@(v) BarrierImpliedVolObjFun(v,S0,Strike,r,q,Mat,BarrierPrice(k),Barrier(k),Rebate), start);
        start = 0.16;
        BarrierIV(k,2) = fminsearch(@(v) BarrierImpliedVolObjFun(v,S0,Strike,r,q,Mat,BarrierPrice(k),Barrier(k),Rebate), start);
    end
    if Barrier(k) == 555
        start = 0.06;
        BarrierIV(k,1) = fminsearch(@(v) BarrierImpliedVolObjFun(v,S0,Strike,r,q,Mat,BarrierPrice(k),Barrier(k),Rebate), start);
        start = 0.11;
        BarrierIV(k,2) = fminsearch(@(v) BarrierImpliedVolObjFun(v,S0,Strike,r,q,Mat,BarrierPrice(k),Barrier(k),Rebate), start);
    end
    if Barrier(k) >= 560
        start = 0.1;
        BarrierIV(k,1) = fminsearch(@(v) BarrierImpliedVolObjFun(v,S0,Strike,r,q,Mat,BarrierPrice(k),Barrier(k),Rebate), start);
    end
end

%% Table 4 of Andersen and Brotherton-Ratcliff
fprintf('--------------------------------------------------------------\n')
fprintf('Table 4 of Andersen and Brotherton-Ratclife \n')
fprintf('%s grid \n',GridType)
fprintf('Using price grid of %4.0f time grid of %4.0f \n',NS, nT+1)
fprintf('--------------------------------------------------------------\n')
fprintf('Barrier    Price   TruePrice DollarError ImpVol_1   ImpVol_2 \n')
fprintf('--------------------------------------------------------------\n')
for k=1:length(Barrier)
    fprintf(' %3.0f   %10.4f  %8.4f  %8.4f  %8.4f   %8.4f \n',...
        Barrier(k),BarrierPrice(k),TruePrice(k),DollarError(k),BarrierIV(k,1),BarrierIV(k,2))
end
fprintf('--------------------------------------------------------------\n')

